/* S.R. Delwiche, USDA/ARS/BA/ANRI/ISL, Building 303, BARC-East, Beltsville, MD 20705-2350 USA, delwiche@ba.ars.usda.gov
   Program filename: forage_discr_six_class_on_SG2D_PCA_cal_xval_val.sas
   Date last modified: October 25, 2006
   Purpose: Reads in Jim's 241-member forage dataset, divides it into the 161/80 split that was done
            for PLS forage concentration calibrations, then does PCA on cal set and forms dendogram.
            Reclassifies the 15 classes into 6, based on the dendogram of the PC scores from running
            the program 'forage_dendro_on_SG2D_PCA_cal_val.sas'.                                                                    */

/* Note: This program is not set up to do looping.  Therefore, the user has to explicitly change the values of certain variables in
         order to apply a Savitzky-Golay convolution.  The changes must be made in the lines that have the comment "change" .  The
         program is not set up to perform MSC or SNV scatter corrections nor variance scaling.                                       */ 

/* ----- Macro variables that seldom need changing ----- */
%Let maxs = 700; /* maximum number of spectral data points */
%Let maxb = 2800; /* 4*maxs */

/* ----- Macro variables that specify input files ----- */
%Let specfile    =d:\pls_with_validation\jim_discrim\ALL241II.SPC;  /* must be Grams multifile (newstyle, even x spacing, constant #pts per spectrum) */
%Let chemfile    =d:\pls_with_validation\jim_discrim\ALL241II_2char_class.prn;
%Let fiberfile   =d:\PLS_with_validation\Jim_Discrim\COMPLETE.PRN;


/* ----- Macro variables that truncate the spectral region before any other preprocessing ----- */
%Let  leftcut = 0; /* the number of spectral points on left end to cut */
%Let rightcut = 0; /* ditto for right end */


/* ........................................... BEGIN MACRO SECTION  .......................................... */

/* Macro that reads a single newstyle Grams multifile (evenly spaced Y values) */

%MACRO getspec;

  data specdata;
  array x(&maxb) x1-x&maxb;
  array y(&maxs) y1-y&maxs;
  infile "&specfile" recfm=n;
  input @2 spcversion ibr1.;
  if spcversion < 0 then spcversion = spcversion + 256;
  if spcversion ^= 75 then abort;
  input @5 numpoints1 ibr1. numpoints2 ibr1. @24 +1 numspec1 ibr1. numspec2 ibr1. @512;
  if numpoints1 < 0 then numpoints1 = numpoints1 + 256;
  if numpoints2 < 0 then numpoints2 = numpoints2 + 256;
  numpoints = numpoints2*256 + numpoints1;

  leftedge = &leftcut + 1;
  attrib charleft format = $CHAR.;
  charleft = leftedge;
  charleft = compress(charleft);
  call symput('leftspec',charleft);
  rightedge = numpoints - &rightcut;
  attrib charright format = $CHAR.;
  charright = rightedge;
  charright = compress(charright);
  call symput('rightspec',charright);
  newnumpoints = numpoints - &leftcut - &rightcut;
  attrib charnewnumpoints format = $CHAR.;
  charnewnumpoints = newnumpoints;
  call symput('newpoints',charnewnumpoints);
  if numspec1 < 0 then numspec1 = numspec1 + 256;
  if numspec2 < 0 then numspec2 = numspec2 + 256;
  numrecords = numspec2*256 + numspec1;
  attrib charrec format = $CHAR.;
  charrec = numrecords;
  charrec = compress(charrec);
  call symput('nrec',charrec);
  do m=1 to numrecords;
    if m = 1 then input +2 mult ibr1. (junk1-junk6) (ibr1.) +24;
    else input +1 mult ibr1. (junk1-junk6) (ibr1.) +24;
    if mult < 0 then mult = mult + 256;
    if mult >= 128 then mult = mult - 256;
    mult = 2 ** (mult - 32);
    do n = 1 to numpoints*4;
      input x(n) ibr1.;
    end;
    do i=4 to numpoints*4 by 4;
      do j = 0 to 3 by 1;
        if x(i-j) < 0 then x(i-j) = x(i-j) + 256;
      end;
      sign = 1;
      if x(i) >= 128 then sign = -1;
      k = i / 4;
      y(k) = (  x(i-3)
              + x(i-2)*256
              + x(i-1)*65536
              + x(i)*16777216 );
      if sign < 0 then y(k) = -(4294967296 - y(k));
      y(k) = mult * y(k);
    end;
    output;
  end;
  stop;
  format y1-y&maxs f13.10;
  keep y1-y&maxs numpoints numrecords;
  run;



%MEND getspec;


%MACRO getchem;

data forageclass;
infile "&chemfile";
input fname $ foragetype $;
classorder = _n_;
if foragetype in('SH') then newtype = 'SH';
else if foragetype in('RH') then newtype = 'RH';
else if foragetype in('CC') then newtype = 'CC';
else if foragetype in('PH') then newtype = 'PH';
else if foragetype in('CS','BS','FS','II','MH','RC','SS','TY','WS') then newtype = 'XX';
else if foragetype in('AL','OG') then newtype = 'YY';
else newtype = 'ZZ';
/* Note that you could collapse some of the categories for foragetype with
   a series of stmts similar to "if foragetype in ('xx','yy') then foragetype = 'zz';"  */
run;


proc sort data = forageclass;
by fname;
quit;

data fiberinfo;
infile "&fiberfile";
input fname $ / chem1-chem6;
fiberorder = _n_;
if _n_ <= 161 then calval = 'cal';
else calval = 'val';
drop chem1-chem6;
run;

proc sort data = fiberinfo;
by fname;
quit;

data combin;
merge forageclass (in = a) fiberinfo (in = b);
by fname;
run;

proc sort data = combin;
by classorder;
quit;


%MEND getchem;

/* -----------------------------------------  END MACRO SECTION  ---------------------------------------------------------- */



/* --------------------------------------------  MAIN PROGRAM  ------------------------------------------------------------ */


options pageno = 1;

%getspec; /* macro call to get spectral data */

%getchem; /* macro call to get classification data */


data foragewithclass;
merge specdata combin;
run;

data forage;
set foragewithclass;
mergevariable = 'a';
run;

data savgol;
infile "d:\pls_with_validation\jim_discrim\savgolgap.prn";
do table = 1 to 10;
  do row = 1 to 14;
    input c1-c11;
    output;
  end;
end;
run;

data chosentable;
set savgol;
if table ^= 6 then delete;                                                                      /* change */
run;

proc transpose data=chosentable out=t_chosentable;
var c1-c11;
quit;

data convolute;
set t_chosentable;
table = 6;                                                                                      /* change */  
SG_pts = 11;                                                                                    /* change */
rename col1=r1 col2=r2 col3=r3 col4=r4 col5=r5 col6=r6 col7=r7 col8=r8 col9=r9
       col10=r10 col11=r11 col12=r12 col13=r13 col14=r14;
if table = 3 or table = 4 or table = 5 or table = 9 then sign = -1; else sign = 1;
mergevariable = 'a';
keep sign SG_pts table col1-col14 mergevariable;
if _n_ ^= 13.5 - SG_pts/2 then delete;
run;

data deriv;
merge forage convolute;
by mergevariable;
array y(1:700) y1-y700;
array r(14) r1-r14;
array z(1:700) z1-z700;
newleft = (SG_pts+1)/2; 
newright = 700 - (SG_pts-1)/2;
do a = newleft to newright;
  sum = 0.0;
  do aa = (14 - (SG_pts+1)/2) to 13;
    sum = sum +        r(aa) * y(a-(13-aa));
  end;
  do aa = (14 - (SG_pts+1)/2) to 12;
    sum = sum + sign * r(aa) * y(a+(13-aa));
  end;
  z(a) = sum / r(14);
end;
keep z1-z700 SG_pts table newleft newright newtype mergevariable calval;
run;

data justcal;
set deriv;
if calval = 'cal';
run;

data justval;
set deriv;
if calval = 'val';
run;


title1 "Jim Reeves' forage data.  After dendogram grouping.";
title2 "LDA analysis on PCA (PCs 1 thru 10) of 11-pt SG2D of NIR spectral data.";               /* change */
title3 "Cal and X-val on 161, val on 80. Collapsed 15 classes into 6.";
title4 "Program is 'forage_discr_six_class_on_SG2D_PCA_cal_xval_val.sas'. Output file = none.";
title5 "Class 'XX' contains the following: CS, BS, FS, II, MH, RC, SS, TY, and WS.";
title6 "Class 'YY' contains the following: AL and OG.";

proc princomp data = justcal cov noprint N=12 outstat = xvalbasis out=pcxval;
var z6-z695;                                                                                    /* change */ 
quit;

proc sort data = pcxval;
by newtype;
quit;

proc score data = justval score = xvalbasis out=pctest;
var z6-z695;                                                                                    /* change */
quit;

proc sort data = pctest;
by newtype;
quit;


proc discrim data=pcxval testdata = pctest method=normal pool=yes crossvalidate outcross = crossvalpreds testout = outtest;
class newtype;
var prin1 prin2 prin3 prin5 prin8 prin4 prin7 prin9 prin10 prin6;
quit;


